/* **********************************************************************
**
**  Copyright (C) 2003  Jesper Hansen <jesperh@telia.com> and 
**			Romuald Bialy (MIS) <romek_b@o2.pl>.
**
*************************************************************************
**
**   This file is part of the yampp system.
**
**  This program is free software; you can redistribute it and/or
**  modify it under the terms of the GNU General Public License
**  as published by the Free Software Foundation; either version 2
**  of the License, or (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software Foundation, 
**  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
**
*********************************************************************** */

#define _SFR_ASM_COMPAT 1

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "atapi.h"
#include "delay.h"
#include "mem.h"
#include "uart.h"

static void SetAddress(u08 addr);
static u08  ReadBYTE(u08 addr);
static void WriteBYTE(u08 addr, u08 dat);
static void CheckBusy(void);
static u08  CheckDataReq(void);
static u08  GetDataFromCD(u08 *buffer);

extern bool bPlaying;

#define MINIBUFFERSIZE 128
u08 minibuffer[MINIBUFFERSIZE];

//----------------------------------------------------------------------------
// Init ATA
//----------------------------------------------------------------------------
void ATA_Init(void)
{
  	WriteBYTE(CMD_A | CMD_DEV_HEAD, DH_CHS0);	// Set Drive/Head register,
							// ie. select drive 0
	delay10us();
	CheckBusy();					// Wait for !BUSY
}


//----------------------------------------------------------------------------
// ATAPI Reset device
//----------------------------------------------------------------------------
void ATA_SW_Reset(void)
{
	delay10us();
	WriteBYTE(CMD_A | CMD_COMMAND, ATAPI_RESET);	// ATAPI Soft Reset command
	delay10us();
	CheckBusy();
	delayms(100);
}

//----------------------------------------------------------------------------
// Clear Packet command buffer;
//----------------------------------------------------------------------------
void ATAPI_ClearBuf(void)
{
	register u08 i;
	for (i=0; i<12; i++)
		minibuffer[i] = 0;
}

// ===========================================================================

// ***************************************************************************
// 				AUDIO CD FUNCTIONS
// ***************************************************************************

//----------------------------------------------------------------------------
// Init Audio, get TOC
//----------------------------------------------------------------------------
u08 ATAPI_DA_GetToc(void)
{
	u08 err;
	DIR_CACHE[5] = 0xFF;				// clear audio toc marker
	ATAPI_ClearBuf();				// clear command buffer
	minibuffer[0]=0x43;				// READ TOC command
	minibuffer[1]=0x02;				// MSF Format
	minibuffer[7]=0x08;				// 2 kB
	err = GetDataFromCD(DIR_CACHE);			// Get data from CD

	if ((DIR_CACHE[5] & 0x0f) > 3)			// if not audio CD
		err = CD_STAT_NOT_READY;		// return no audio CD

	return (err);
}

//----------------------------------------------------------------------------
// get M for Track #
//----------------------------------------------------------------------------
u08  ATAPI_DA_GetM(u08 Track)
{
	return DIR_CACHE[(Track * 8) + 9];		// calculate
}

//----------------------------------------------------------------------------
// get S for Track #
//----------------------------------------------------------------------------
u08  ATAPI_DA_GetS(u08 Track)
{
	return DIR_CACHE[(Track * 8) + 10];		// calculate
}

//----------------------------------------------------------------------------
// get F for Track #
//----------------------------------------------------------------------------
u08  ATAPI_DA_GetF(u08 Track)
{
	return DIR_CACHE[(Track * 8) + 11];		// calculate
}

//----------------------------------------------------------------------------
// get total # of tracks on the CD
//----------------------------------------------------------------------------
u08  ATAPI_DA_GetTrackN(void)
{
	return(((DIR_CACHE[0]<<8) + DIR_CACHE[1] - 4) / 8);
}

//----------------------------------------------------------------------------
// get total disc time (in seconds)
//----------------------------------------------------------------------------
u16  ATAPI_DA_GetDiscTime(void)
{
	u16 m = ATAPI_DA_GetM(ATAPI_DA_GetTrackN());
	u16 s = ATAPI_DA_GetS(ATAPI_DA_GetTrackN());
	return(60*m + s - 2);
}

//----------------------------------------------------------------------------
// Play From M:S:F
//----------------------------------------------------------------------------
u08  ATAPI_DA_PlayMSF(u08 m, u08 s, u08 f)
{
	ATAPI_ClearBuf();
	minibuffer[0] = 0x47;					// Command code (PLAY)
	minibuffer[3] = m;					// start M
	minibuffer[4] = s;					// start S
	minibuffer[5] = f;					// start F
	minibuffer[6] = ATAPI_DA_GetM(ATAPI_DA_GetTrackN());	// end M
	minibuffer[7] = ATAPI_DA_GetS(ATAPI_DA_GetTrackN());	// end S
	minibuffer[8] = ATAPI_DA_GetF(ATAPI_DA_GetTrackN());	// end F
	return ATAPI_Command();					// 0 if OK, else ERROR
}


//----------------------------------------------------------------------------
// Play FromTrack
//----------------------------------------------------------------------------
u08  ATAPI_DA_Play(u08 Track)
{
	return ATAPI_DA_PlayMSF(ATAPI_DA_GetM(Track), ATAPI_DA_GetS(Track), 
				ATAPI_DA_GetF(Track));
}

//----------------------------------------------------------------------------
// Stop Playing
//----------------------------------------------------------------------------
u08  ATAPI_DA_Stop(void)
{
	ATAPI_ClearBuf();
	minibuffer[0] = 0x4E;			// Command code (STOP)
	return ATAPI_Command();			// 0 if OK, else ERROR
}

//----------------------------------------------------------------------------
// Pause, resume (bResume == true) playing
//----------------------------------------------------------------------------
u08  ATAPI_DA_PauseResume(bool bResume)
{
	ATAPI_ClearBuf();
	minibuffer[0] = 0x4B;			// Command code (PAUSE)
	minibuffer[8] = bResume;		// 0 -> pause, 1 -> resume
	return ATAPI_Command();			// 0 if OK, else ERROR
}

//----------------------------------------------------------------------------
// Search function, skipping (CDDA_SCAN_STEP) seconds. 
// bDir=0 -> forward, bDir=1 -> backward
//----------------------------------------------------------------------------
void  ATAPI_DA_Scan(bool bDir)
{
	u08 err;
	if (ATAPI_DA_GetSubC() == 0)
	{
		u08 m = BUFFER1P[9];
		u08 s = BUFFER1P[10];
		if (bDir)
		{
			s -= CDDA_SCAN_STEP;
			if(((s08)s < 0) && m > 0)
			{
				s += 60;
				m--;
			}
			if ((s16)(60*(u16)m+s) < 2)
				return;

		}
		else
		{
			s += CDDA_SCAN_STEP;
			if(s > 59)
			{
				s -= 60;
				m++;
			}
			if ((s16)(60*(u16)m+s) > (s16)ATAPI_DA_GetDiscTime())
				return;

		}
		do{
			err = ATAPI_DA_PlayMSF(m, s, 0);
			delayms(100 * DELAY_TIME);
			err |= ATAPI_DA_GetSubC();
		} while (err);
	}
}

//----------------------------------------------------------------------------
// Get SubChannel Data to BUFFER1 memory (buffers is not used in CDDA mode)
//----------------------------------------------------------------------------
u08  ATAPI_DA_GetSubC(void)
{
	u08 err;
	CheckBusy();
	BUFFER1P[1] = 0xFF;
	ATAPI_ClearBuf();
	minibuffer[0] = 0x42;			// Command code (Read SUBCHANNEL)
	minibuffer[1] = 0x02;
	minibuffer[2] = 0x40;
	minibuffer[3] = 0x01;			// 01 - current position
	minibuffer[7] = 0x08;			// 2 kB
	err = GetDataFromCD(BUFFER1P);		// Get data from CD
	delayms(2);

	if(BUFFER1P[1] == 0x11)
		return 0;
	else
		return (BUFFER1P[1]);
}


//----------------------------------------------------------------------------
// Get Current Time
//----------------------------------------------------------------------------
u16  ATAPI_DA_GetCurTime(void)
{
	return (60*(BUFFER1P[13]) + BUFFER1P[14]);
}

//----------------------------------------------------------------------------
// Get Current Track Number
//----------------------------------------------------------------------------
u08  ATAPI_DA_GetCurTrack(void)
{
	return (BUFFER1P[6] - 1);
}



// ***************************************************************************
// 			DATA CD AND GEBERAL FUNCTIONS
// ***************************************************************************

//----------------------------------------------------------------------------
// Read one 2048 bytes sector in LBA mode from CD
// Returns 0 if no error detected
//----------------------------------------------------------------------------
u08 ATAPI_ReadLBA(u32 LBA, u08 *Buffer)
{
	union u32convert LBAC;
	LBAC.value = LBA;				// Prepare parameters...
	ATAPI_ClearBuf();
	minibuffer[0]=ATAPI_READ12;			// Read(12) Command
	minibuffer[2]=LBAC.bytes.byte4;
	minibuffer[3]=LBAC.bytes.byte3;
	minibuffer[4]=LBAC.bytes.byte2;
	minibuffer[5]=LBAC.bytes.byte1;
	minibuffer[9]=1;	
	return (GetDataFromCD(Buffer));
}

extern u08 cdda;

u08 GetDataFromCD(u08 *Buffer)
{
	asm volatile ("push r24");
	asm volatile ("push r25");			// save buffer address
	u08 i,s;
	u08 j=0;

CMDIN:	i = 0;
	while(ATAPI_Command() && (++j < 10)); // PRINT("\r\nCmdErr");	// Issue ATAPI Packet command
	delayms((cdda) ? 10 : 3);

	if (++j > 10)
	{
		ATA_SW_Reset();
//		PRINT("\r\nATAPI Reset");
		delayms(10);
		j = CD_STAT_UNDEF_ERROR;
		goto EXIT;					// Too more errors > EXIT
	}

	if ((s=CheckDataReq()))					// If error
	{
		i = ReadBYTE(CMD_A | CMD_ERROR);
//		PRINT("\r\nATAPI Error: "); UART_Printfu08(s); UART_SendByte(32);
//		UART_Printfu08(i); UART_SendByte(32); UART_Printfu08(j);
		if (i == 0x20)
		{
			j = CD_STAT_NODISK;				// Media not present > EXIT
			goto EXIT;
		}

		if (i==0xD0 && j>0)
		{
			j--;
			goto CMDIN;
		}

		if (s == 2)
			delayms(10);

		goto CMDIN;
	}
	cli();		

	i = (minibuffer[0] == ATAPI_REQUEST_SENSE) ? 1 : BPS/MINIBUFFERSIZE;

	asm volatile ("pop r27");				// restore buffer address to reg X
	asm volatile ("pop r26");
	do							// cycle sector by minibuffers 
	{
		if(i-- == 0)					// Check count of sended 128 bytes blocks
			return(1);				// CD-ROM try send to more data

		SetAddress(CMD_A + CMD_DATA);			// setup addressing and chip selects
		outp(0x00, DDRA);				// port A as input
		outp(0x00, DDRC);				// port C as input
		outp(0xff, PORTA);				// activate Pullups
		outp(0xff, PORTC);				// activate Pullups

		asm volatile ("ldi r30,lo8(minibuffer)");
		asm volatile ("ldi r31,hi8(minibuffer)");
		j = MINIBUFFERSIZE;
		do						// cycle minibuffer by words
		{
			cbi(PORTB, 1);				// set DIOR lo
			j -= 2;					// nop replacement, allow pin change
			inp(PINA);				// read lo byte	
			asm volatile ("st Z+,r24");		// store to minibuf
			inp(PINC);				// read hi byte
			asm volatile ("st Z+,r24");		// store to minibuf
			sbi(PORTB, 1);				// set DIOR hi
		} while (j > 0);

		sbi(MCUCR, SRE);				// enable ExtRAM

		asm volatile ("ldi r30,lo8(minibuffer)");
		asm volatile ("ldi r31,hi8(minibuffer)");
		j = MINIBUFFERSIZE;
		do
		{
			asm volatile ("ld r24,Z+");
			asm volatile ("st X+,r24");
		} while (--j > 0);

	} while ((ReadBYTE(CMD_A | CMD_STATUS) & (SR_BUSY | SR_DRQ)) == SR_DRQ);

	ReadBYTE(CTRL_A | CTRL_ALT_STAT);
	sei();							// enable interrupt
	CheckBusy();
	return(ReadBYTE(CMD_A | CMD_STATUS) & SR_ERR);		// return 0 if OK, 1 if error

EXIT:	asm volatile ("pop r25");
	asm volatile ("pop r25");
	return (j);
}

//----------------------------------------------------------------------------
// Select address and CS signals
//----------------------------------------------------------------------------
void SetAddress(u08 addr)
{
	cbi(MCUCR, SRE);		// Disable ExtRAM
	outp(0xff, DDRA);
	outp(addr, PORTA);
	sbi(PORTE,PE1);
	cbi(PORTE,PE1);
}


//----------------------------------------------------------------------------
// Read data BYTE from Drive
//----------------------------------------------------------------------------
u08 ReadBYTE(u08 addr)
{
register u08 tmp;
	SetAddress(addr);
	outp(0x00, DDRA);		// port A as input
	outp(0x00, DDRC);		// port C as input
	outp(0xff, PORTA);		// activate Pullups
	asm volatile ("nop");		// allow pin change
	cbi(PORTB, 1);			// set DIOR lo
	asm volatile ("nop");		// allow pin change
	asm volatile ("nop");		// allow pin change
	tmp = inp(PINA);		// read byte
	sbi(PORTB, 1);			// set DIOR hi
	asm volatile ("nop");		// allow pin change
	sbi(MCUCR, SRE);		// enable RAM
  	return tmp;
}


//----------------------------------------------------------------------------
// Write data BYTE to Drive
//----------------------------------------------------------------------------
void WriteBYTE(u08 addr, u08 dat)
{
  	SetAddress(addr);
	outp(0xff, DDRC);		// port C as output
	outp(dat, PORTA);		// write byte
	outp(0,   PORTC);
	asm volatile ("nop");		// allow pin change
	cbi(PORTB, 0);			// set DIOW lo
	asm volatile ("nop");		// allow pin change
	asm volatile ("nop");		// allow pin change
	sbi(PORTB, 0);			// set DIOW hi
	asm volatile ("nop");		// allow pin change
	sbi(MCUCR, SRE);		// enable RAM
}

//----------------------------------------------------------------------------
// Send the PACKET command to CD-ROM and 12 command data bytes from minibuffer
//----------------------------------------------------------------------------
u08 ATAPI_Command()
{
	register u08 i = 0;
	CheckBusy();
	WriteBYTE(CMD_A | CMD_FEATURES , 0);
	WriteBYTE(CMD_A | CMD_CYL_HIGH , 8);			// Max transfer length
	WriteBYTE(CMD_A | CMD_CYL_LOW ,  0);			// = 2048 bytes
	WriteBYTE( CMD_A | CMD_COMMAND, ATAPI_PACKET );		// Write command Packet
	delay10us();
	CheckBusy();
	cli();
  	SetAddress(CMD_A | CMD_DATA);
	outp(0xff, DDRC);					// port C as output
	for(i=0; i<12;)
	{
		outp(minibuffer[i++], PORTA);			// write lo byte
		outp(minibuffer[i++], PORTC);			// write hi byte
		asm volatile ("nop");				// allow pin change
		cbi(PORTB, 0);					// set DIOW lo
		asm volatile ("nop");				// allow pin change
		asm volatile ("nop");				// allow pin change
		sbi(PORTB, 0);					// set DIOW hi
		asm volatile ("nop");				// allow pin change
	}
	sbi(MCUCR, SRE);					// enable RAM
	delay10us();
	ReadBYTE(CTRL_A | CTRL_ALT_STAT);
	sei();
	CheckBusy();
	return (ReadBYTE(CMD_A | CMD_STATUS) & SR_ERR);		// Return command status
}


//----------------------------------------------------------------------------
// Check CD-ROM status
//----------------------------------------------------------------------------
u08 ATAPI_CheckStatus(void)
{
	ATAPI_ClearBuf();
	ATAPI_Command();
	delayms(1);
	
	ATAPI_ClearBuf();
	minibuffer[0] = ATAPI_REQUEST_SENSE;		// Request Sense
	minibuffer[4] = 18;				// Allocation length
	GetDataFromCD(SENSE_BUF);
/*
		u08 i;
		for (i=0; i<19; i++)
		{
			UART_Printfu08(SENSE_BUF[i]);
			UART_SendByte(32);
		}
		EOL();
*/
	switch (SENSE_BUF[12])
	{
		default:	return CD_STAT_UNDEF_ERROR;
		case 0x29:
		case 0x04:	return CD_STAT_NOT_READY;
		case 0x3A:	return CD_STAT_NODISK;
		case 0x00:	return CD_STAT_OK;
	}
	return CD_STAT_OK;
}

//----------------------------------------------------------------------------
// Slow down CD-ROM unit
//----------------------------------------------------------------------------
void SlowDown()
{
#ifdef READ_SPEED
	ATAPI_ClearBuf();
	minibuffer[0] = ATAPI_SLOWDOWN;
	minibuffer[2] = (READ_SPEED)>>8;
	minibuffer[3] = (u08)(READ_SPEED);
	ATAPI_Command();
#endif
}

//----------------------------------------------------------------------------
// Load or Unload CD
//----------------------------------------------------------------------------
void CD_Cmd(u08 Funct)
{
	ATAPI_ClearBuf();
	minibuffer[0] = ATAPI_START_STOP_UNIT;
	minibuffer[1] = 0x01;			// Immed = 1
	minibuffer[4] = Funct;
	ATAPI_Command();
}

//----------------------------------------------------------------------------
// Check ATA busy status
//----------------------------------------------------------------------------
void CheckBusy(void)
{
	u16 tout = 0;
	WDR;
	while (((ReadBYTE(CMD_A | CMD_STATUS) & SR_BUSY) == SR_BUSY) && (++tout < 1000))
		delay10us();
}

//----------------------------------------------------------------------------
// Check ATA data request status, 0=OK, 1=ERROR, 2=Timeout
//----------------------------------------------------------------------------
u08 CheckDataReq(void)
{
	u16 timeout = 0;
	u16 maxtout = 500 + (50000 * (bPlaying == 0));
	register u08 stat;
	WDR;
	while(1)
	{
		stat = ReadBYTE(CMD_A | CMD_STATUS);
		if (stat & SR_ERR) 
			return 1;
		if (++timeout > maxtout) 
			return 2;
		if ((stat & (SR_BUSY | SR_DRQ)) == SR_DRQ)
			return 0;
		delay10us();
	}
}
